r"""
Plotting functions

.. autosummary::
   :toctree: generated/

   compare_groups_of_contacts
   plot_matrix
   plot_unified_freq_dicts
   plot_unified_distro_dicts
   compare_violins
   color_dict_guesser

"""

from .plots import *


r"""
Tools for reading and manipulating sites.

.. autosummary::
   :toctree: generated/
   :nosignatures:

   x2site
   sites_to_res_pairs

Sites are user-defined collections of contacts.
They can be constructed by hand or read from
plain ascii files:

>>> cat site.dat
# contacts to look at :
L394-K270
D381-Q229
Q384-Q229
R385-Q229
D381-K232
Q384-I135

For unformatted ascii files (.dat, .txt etc),
if the first line starts with "#", it will be used
as name for the site, otherwise the filename
will be used. Also, any lines starting
with "#" will be ignored

The ascii-files can be annotated if the `JSON <https://www.json.org/json-en.html>`_
format is used:

>>> cat site.json
{"name":"interesting contacts",
"pairs": {"AAresSeq": [
        "L394-K270",
        "D381-Q229",
        "Q384-Q229",
        "R385-Q229",
        "D381-K232",
        "Q384-I135"
        ]}}

Sites can also be used inside the Python script/session
by using dictionaries, e.g.::
   my_site = {"name":"interesting contacts",
              "pairs":{"AAresSeq":[
                        "L394-K270",
                        "D381-Q229",
                        "Q384-Q229",
                        "R385-Q229",
                        "D381-K232",
                        "Q384-I135"
                        ]}}
Or ::
   my_site = {"name":"interesting contacts",
              "pairs":{"AAresSeq":[
                        ["L394","K270"],
                        ["D381","Q229"],
                        ["Q384","Q229"],
                        ["R385","Q229"],
                        ["D381","K232"],
                        ["Q384","I135"]
                        ]}}

You can specify the "pairs" as "AAresSeq" (like above)
or as zero-indexed residue serial indices using "residx"::
   my_site = {"name":"interesting contacts",
              "pairs":{"residx":[
                        "353-972",
                        "340-956",
                        "343-956",
                        "344-956",
                        "340-959",
                        "343-865"
                        ]}}

In this last case, you can also directly use pairs of integers,
instead of strings::
  my_site = {"name":"interesting contacts",
             "pairs":{"residx":[
                      [353,972],
                      [340,956],
                      [343,956],
                      [344,956],
                      [340,959],
                      [343,865],
                      ]}}

Using "AAresSeq" makes site-definitions
portable of across topologies:
e.g. "L394" will be picked regardless of the actual
residue index. This can be problematic if
there's duplicates (e.g. "ALA150" appears two times
in 3SN6). "residx" avoid this, but first the user
needs to find out which residue index they are interested
in. mdciao offers several ways to do this. From the CLI
you can use:

>>> mdc_residue.py ALA150 3SN6.pdb

From the API, you can use:

>>> mdciao.cli.residue_selection("ALA150",geom);
   Using method 'lig_resSeq+' these fragments were found
   fragment      0 with  349 AAs     THR9           (   0) -   LEU394           (348 ) (0)  resSeq jumps
   fragment      1 with  340 AAs     GLN1           ( 349) -   ASN340           (688 ) (1)
   fragment      2 with  217 AAs     ASN5           ( 689) -  ALA1160           (905 ) (2)  resSeq jumps
   fragment      3 with  284 AAs    GLU30           ( 906) -   CYS341           (1189) (3)  resSeq jumps
   fragment      4 with  128 AAs     GLN1           (1190) -   SER128           (1317) (4)
   fragment      5 with    1 AAs  P0G1601           (1318) -  P0G1601           (1318) (5)
   0.0)       ALA150 in fragment 0 with residue index 113
   0.1)       ALA150 in fragment 0 with residue index 1026
   Your selection 'ALA150' yields:
      residue      residx    fragment      resSeq      GPCR        CGN
       ALA150         113           0        150       None       None
       ALA150        1026           0        150       None       None

Check :obj:`mdciao.cli.residue_selection` for more.


"""
from .siteIO import *


r"""

Container for other sub-modules with lower-level functions.
They are exposed, tested, and documented, so feel free to
use them in your workflow as needed.

.. currentmodule:: mdciao.utils

.. autosummary::
   :nosignatures:
   :toctree: generated/

   residue_and_atom
   bonds
   str_and_dict
   COM
   lists
   contact_matrix
   sequence

"""

from . import str_and_dict
from . import bonds
from . import residue_and_atom
from . import COM
from . import lists
from . import contact_matrix
from . import sequence


r"""

Computation, bookkeeping, and manipulation
of residue-residue contacts.

The methods wrap around a modified version
of `mdtraj.compute_contacts` to extract
residue-residue distances. The modifications
consist in including the indices of the closest
atom-pairs in the returned values.

The classes contain and abstract these distances into
sense-making groups, offering methods to operate directly
on all residue-residue distances.

It is recommended to use higher-level methods of the API,
like those exposed by :obj:`mdciao.cli` to create
:obj:`ContactPair` or :obj:`ContactGroup` objects and
then use the methods stated above, but of course experienced
users can instantiate them directly.

.. currentmodule:: mdciao.contacts


Classes
=======

.. autosummary::
    :nosignatures:
    :toctree: generated

    Residues
    ContactPair
    ContactGroup
    GroupOfInterfaces

Functions
=========

.. autosummary::
    :nosignatures:
    :toctree: generated/

    per_traj_ctc
    trajs2ctcs
    select_and_report_residue_neighborhood_idxs

"""
from .contacts import *
from .contacts import _linear_switchoff



##############################################################################
#    This file is part of mdciao.
#
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

##############################################################################
#    Please note: The following method "from_dataframe" is a
#    modified version of the one found in the original MDTraj
#    Python Library, whose original authors and copyright
#    holders are listed below.

#    The modifications consist in a bugfix that adds a third condition
#    to force new residue creation. A pull request has been already
#    opened upstream and hopefully soon this file won't be needed
#
##############################################################################


##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Christian Schwantes
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

# opened issues for this:
# https://github.com/mdtraj/mdtraj/issues/1569
# https://github.com/mdtraj/mdtraj/pull/1740
# Further reading
# https://www.oreilly.com/library/view/understanding-open-source/0596005814/ch03.html
# https://tldrlegal.com/license/gnu-lesser-general-public-license-v3-(lgpl-3)
# http://oss-watch.ac.uk/resources/lgpl
# https://www.gnu.org/licenses/gpl-faq.html#AllCompatibility

##############################################################################
# Imports
##############################################################################

from __future__ import print_function, division

import itertools
import numpy as np
import os
import xml.etree.ElementTree as etree
from collections import namedtuple

from mdtraj.core import element as elem
from mdtraj.core.residue_names import (_PROTEIN_RESIDUES, _WATER_RESIDUES,
                                       _AMINO_ACID_CODES)
from mdtraj.core.selection import parse_selection
from mdtraj.utils import ilen, import_, ensure_type
from mdtraj.utils.six import string_types
from mdtraj.utils.singleton import Singleton

from mdtraj.core.topology import Atom, float_to_bond_type

def from_dataframe(atoms, bonds=None):
    """Create a mdtraj topology from a pandas data frame

    Parameters
    ----------
    atoms : pandas.DataFrame
        The atoms in the topology, represented as a data frame. This data
        frame should have columns "serial" (atom index), "name" (atom name),
        "element" (atom's element), "resSeq" (index of the residue)
        "resName" (name of the residue), "chainID" (index of the chain),
        and optionally "segmentID", following the same conventions
        as wwPDB 3.0 format.
    bonds : np.ndarray, shape=(n_bonds, 4) or (n_bonds, 2), dtype=float, Optional
        The bonds in the topology, represented as a n_bonds x 4 or n_bonds x 2
        size array of the indices of the atoms involved, type, and order of each
        bond, represented as floats. Type and order are optional. Specifying bonds
        here is optional. To create standard protein bonds, you can use
        `create_standard_bonds` to "fill in" the bonds on your newly created
        Topology object, although type and order of bond are not computed if
        that method is used.

    See Also
    --------
    create_standard_bonds
    """
    pd = import_('pandas')

    if bonds is None:
        bonds = np.zeros([0, 4], dtype=float)

    for col in ["name", "element", "resSeq",
                "resName", "chainID", "serial"]:
        if col not in atoms.columns:
            raise ValueError('dataframe must have column %s' % col)

    if "segmentID" not in atoms.columns:
        atoms["segmentID"] = ""

    from mdtraj import Topology
    out = Topology()
    if not isinstance(atoms, pd.DataFrame):
        raise TypeError('atoms must be an instance of pandas.DataFrame. '
                        'You supplied a %s' % type(atoms))
    if not isinstance(bonds, np.ndarray):
        raise TypeError('bonds must be an instance of numpy.ndarray. '
                        'You supplied a %s' % type(bonds))

    if not np.all(np.arange(len(atoms)) == atoms.index):
        raise ValueError('atoms must be uniquely numbered '
                         'starting from zero.')
    out._atoms = [None for i in range(len(atoms))]

    c = None
    r = None
    previous_chainID = None
    previous_resName = None
    previous_resSeq = None
    for atom_index, atom in atoms.iterrows():

        int(atom_index)  # Fixes bizarre hashing issue on Py3K.  See #545

        if atom['chainID'] != previous_chainID:
            previous_chainID = atom['chainID']

            c = out.add_chain()

        if atom['resSeq'] != previous_resSeq or atom['resName'] != previous_resName or c.n_atoms == 0:
            previous_resSeq = atom['resSeq']
            previous_resName = atom['resName']

            r = out.add_residue(atom['resName'], c, atom['resSeq'], atom['segmentID'])

        a = Atom(atom['name'], elem.get_by_symbol(atom['element']),
                 atom_index, r, serial=atom['serial'])
        out._atoms[atom_index] = a
        r._atoms.append(a)

    for bond in bonds:
        ai1 = int(bond[0])
        ai2 = int(bond[1])
        try:
            bond_type = float_to_bond_type(bond[2])
            bond_order = int(bond[3])
            if bond_order == 0:
                bond_order = None
        except IndexError:
            # Does not exist
            bond_type = None
            bond_order = None
        out.add_bond(out.atom(ai1), out.atom(ai2), bond_type, bond_order)

    out._numAtoms = out.n_atoms
    return out



r"""

Helper functions for demos and data download

.. currentmodule:: mdciao.examples

.. autosummary::
   :toctree: generated/
   :nosignatures:

   ContactGroupL394
   Interface_B2AR_Gas
   GPCRLabeler_ardb2_human
   CGNLabeler_3SN6
   notebooks
   fetch_example_data
   filenames
"""
from .examples import ContactGroupL394, Interface_B2AR_Gas, GPCRLabeler_ardb2_human, CGNLabeler_3SN6, filenames, notebooks, fetch_example_data



r"""

Deal with residues, atoms, and their names, mostly.
The function :obj:`residues_from_descriptors` is probably the
most elaborate and most higher-level.

.. autosummary::
   :nosignatures:
   :toctree: generated/


"""
from mdtraj.core.residue_names import _AMINO_ACID_CODES
import mdtraj as _md
from fnmatch import filter as _fn_filter
import numpy as _np
from  pandas import unique as _pandas_unique
from mdciao.utils.lists import in_what_N_fragments as _in_what_N_fragments, force_iterable as _force_iterable
from collections import Counter as _Counter
from pandas import DataFrame as _DF

def residues_from_descriptors(residue_descriptors,
                              fragments, top,
                              pick_this_fragment_by_default=None,
                              fragment_names=None,
                              additional_resnaming_dicts=None,
                              extra_string_info='',
                              just_inform=False,
                              ):
    r"""
    Returns residue idxs based on a list of residue descriptors.

    Fragments are needed to better identify residues. If a residue
    is present in multiple fragments, the user can dis-ambiguate
    or pick all residue idxs matching the :obj:`residue_descriptor`

    Because of this (one descriptor can match more than one residue)
    the return values are not necessarily of the same length
    as :obj:`residue_descriptors`

    Parameters
    ----------
    residue_descriptors: string or list of of strings
        AAs of the form of "GLU30" or "E30" or 30, can be mixed
    fragments: iterable of iterables of integers
        The integers in the iterables of 'fragments'
        represent residue indices of that fragment
    top: :obj:`~mdtraj.Topology`
    pick_this_fragment_by_default: None or integer.
        Pick this fragment without asking in case of ambiguity.
        If None, the user will we prompted
    fragment_names:
        list of strings providing informative names input :obj:`fragments`
    additional_resnaming_dicts : dict of dicts, default is None
        Dictionary of dictionaries. Lower-level dicts are keyed
        with residue indices and valued with additional residue names.
        Higher-level keys can be whatever. Use case is e.g. if "R131"
        needs to be disambiguated bc. it pops up in many fragments.
        You can pass {"GPCR":{895:"3.50", ...} here and that label
        will be displayed next to the residue. :obj:`mdciao.cli`
        methods use this.
    just_inform : bool, default is False
        Just inform about the AAs, don't ask for a selection
    extra_string_info: string with any additional info to be printed in case of ambiguity

    Returns
    -------
    residxs : list
        lists of integers that have been selected
    fragidxs : list
        The list of fragments where the residues are
    """
    residxs = []
    fragidxs = []
    last_answer = None
    first_run = True

    if isinstance(residue_descriptors, (str, int)):
        residue_descriptors = [residue_descriptors]

    for key in residue_descriptors:
        cands = _np.array(find_AA(str(key), top,extra_columns=additional_resnaming_dicts))
        cand_fragments =   _force_iterable(_np.squeeze(_in_what_N_fragments(cands, fragments)))
        # TODO refactor into smaller methods
        if len(cands) == 0:
            print("No residue found with descriptor %s" % key)
            residxs.append(None)
            fragidxs.append(None)
        elif len(cands) == 1:
            if len(cand_fragments)>1:
                raise ValueError("Your fragment definitions overlap, "
                                 "res_idx %u (%s) is found in fragments  %s"%
                                 (cands[0], top.residue(cands[0]), ', '.join([str(fr) for fr in cand_fragments])))
            elif len(cand_fragments)==0:

                raise ValueError("Your fragment definitions do not contain "
                                 "the residue of interest %u (%s)."%
                                 (cands[0], top.residue(cands[0])))
            residxs.append(cands[0])
            fragidxs.append(cand_fragments[0])
            if just_inform:
                istr = residue_line("0.0", top.residue(residxs[-1]), fragidxs[-1], additional_resnaming_dicts, fragment_names=fragment_names)
                print(istr)
        else:
            istr = "Ambiguous definition for residue %s" % key
            istr += extra_string_info
            if not just_inform:
                print(istr)
            cand_chars = _np.hstack([['%s.%u'%(key,ii) for ii in range(n)] for key, n in _Counter(cand_fragments).items()]).tolist()
            for cc, ss, char in zip(cands, cand_fragments, cand_chars):
                istr = residue_line(char, top.residue(cc), ss, additional_resnaming_dicts, fragment_names=fragment_names)
                print(istr)
            if just_inform:
                print()
                residxs.extend([ii for ii in cands if ii not in residxs])
                fragidxs.extend([ii for ii in cand_fragments if ii not in fragidxs])
                continue
            if pick_this_fragment_by_default is None:
                if first_run is True:
                    last_answer = [int(ii) for ii in _np.unique(cand_fragments)][0]
                    choice_str = "choose first candidate fragment [%s]" % last_answer
                    first_run = False
                else:
                    choice_str = "repeat last option [%s]" % last_answer
                prompt = "Input one fragment idx out of %s and press enter (selects all matching residues in that fragment).\n" \
                         "Use one x.y descriptor in case of repeated fragment index.\n" \
                         "Leave empty and hit enter %s" % (
                         [int(ii) for ii in _np.unique(cand_fragments)], choice_str)

                answer = input(prompt)
            else:
                answer = str(pick_this_fragment_by_default)
                print("Automatically picked fragment %u" % pick_this_fragment_by_default)

            if len(answer) == 0:
                answer = last_answer

            if str(answer).isdigit():
                answer = int(answer)
                assert answer in cand_fragments
                idxs_w_answer = _np.argwhere([answer == ii for ii in cand_fragments]).squeeze()
                cands = cands[idxs_w_answer]
            elif '.' in str(answer) and answer in cand_chars:
                idx_w_answer = _np.argwhere([answer == ii for ii in cand_chars]).squeeze()
                answer = cand_fragments[idx_w_answer]
                cands = cands[idx_w_answer]
            else:
                raise ValueError("%s is not a possible answer" % answer)
                # TODO implent k for keeping this answer from now on

            assert answer in cand_fragments, (
                    "Your answer has to be an integer in the of the fragment list %s" % cand_fragments)
            last_answer = answer

            residxs.extend([int(ii) for ii in _force_iterable(cands)])
            fragidxs.extend([int(answer) for __ in _force_iterable(cands)])

    return residxs, fragidxs

def rangeexpand_residues2residxs(range_as_str, fragments, top,
                                 interpret_as_res_idxs=False,
                                 sort=False,
                                 **residues_from_descriptors_kwargs):
    r"""
    Generalized range-expander from residue descriptors.

    Residue descriptors can be anything that :obj:`find_AA` understands.
    Expanding a range means getting "2-5,7" as input and returning "2,3,4,5,7"

    To dis-ambiguate descriptors, a fragment definition and a topology are needed

    TODO
    ----
    Internally, if an int or an iterable of ints is passed, they
    will be string-ified on-the-fly to work with the method "as-is",
    because the method was initially developed to interpret CLI-strings

    Note
    ----
    The input (= compressed range) is very flexible and accepts
    mixed descriptors and wildcards, eg: GLU*,ARG*,GDP*,LEU394,380-385 is a valid range.

    Wildcards use the full resnames, i.e. E* is NOT equivalent to GLU*

    Be aware, though, that wildcards are very powerful and easily "grab" a lot of
    residues, leading to long calculations and large outputs.

    See :obj:`find_AA` for more on residue descriptors

    Parameters
    ----------
    range_as_str : string, int or iterable of ints
    fragments : list of iterable of residue indices
    top : :obj:`~mdtraj.Topology` object
    interpret_as_res_idxs : bool, default is False
        If True, indices without residue names ("380-385") values will be interpreted as
        residue indices, not residue sequential indices
    sort : bool
        sort the expanded range on return
    residues_from_descriptors_kwargs:
        Optional parameters for :obj:`residues_from_descriptors`

    Returns
    -------
    residxs_out = list of unique residue indices
    """
    residxs_out = []
    #print("For the range", range_as_str)
    if not isinstance(range_as_str,str):
        range_as_str = _force_iterable(range_as_str)
        assert all([isinstance(ii,(int,_np.int64)) for ii in range_as_str]),(range_as_str,[type(ii)  for ii in range_as_str])
        range_as_str= ','.join([str(ii) for ii in range_as_str])
    for r in [r for r in range_as_str.split(',') if r!=""]:
        assert not r.startswith("-")
        if "*" in r or "?" in r:
            assert "-" not in r
            filtered = find_AA(r, top, extra_columns= residues_from_descriptors_kwargs.get("additional_resnaming_dicts"))
            if len(filtered)==0:
                raise ValueError("The input range contains '%s' which "
                                 "returns no residues!"%r)
            residxs_out.extend(filtered)
        else:
            resnames = r.split('-')
            is_range = "-" in r
            if interpret_as_res_idxs:
                # TODO clean the double "-" condiditon 
                if is_range:
                    assert len(resnames) == 2
                    for_extending = _np.arange(int(resnames[0]),
                                               int(resnames[-1]) + 1)
                else:
                    for_extending = [int(rr) for rr in resnames]
            else:
                for_extending, __ = residues_from_descriptors(resnames, fragments, top,
                                                            **residues_from_descriptors_kwargs)
                if None in for_extending:
                    for idesc in [int_from_AA_code(str(r)), name_from_AA(r)]:
                        if idesc not in [None,""]:
                            print("Trying with '%s'"%idesc)
                            residues_from_descriptors(idesc, fragments,top, just_inform=True)
                    raise ValueError("The input range of residues contains '%s' which "
                                     "returns an untreatable range %s!\nCheck the above list for help." % (r, for_extending))
                if is_range:  # it was a pair
                    assert len(for_extending)==2
                    for_extending = _np.arange(for_extending[0],
                                               for_extending[-1] + 1)
                else: # it was something else
                    pass

            residxs_out.extend(for_extending)

    if sort:
        residxs_out = sorted(residxs_out)

    residxs_out = _pandas_unique(residxs_out)
    return residxs_out

def int_from_AA_code(key):
    """
    Returns the integer part from a residue name, None if there isn't

    Parameters
    ----------
    key : string
        Residue name passed as a string, example "GLU30"

    Returns
    -------
    int
        Integer part of the residue id, example- 30 if the input is "GLU30"

    """
    try:
        return int(''.join([ii for ii in key if ii.isnumeric()]))
    except ValueError:
        return None

def name_from_AA(key):
    """
    Return the residue name from a string

    Parameters
    ----------
    key : string or obj:`mdtraj.Topology.Residue` object
        Residue name passed as a string, example "GLU30" or as residue object

    Returns
    -------
    name: str
        Name of the residue, like "GLU" for "GLU30" or "E" for "E30"

    """


    return ''.join([ii for ii in str(key) if ii.isalpha()])

def shorten_AA(AA, substitute_fail=None, keep_index=False):
    r"""
    Return the short name of an AA, e.g. TRP30 to W by trying to
    use either the :obj:`mdtraj.Topology.Residue.code` attribute
    or :obj:`mdtraj` internals AA dictionary

    Parameters
    ----------
    AA: :obj:`~mdtraj.Topology.Residue` or a str
        The residue in question

    substitute_fail: str, default is None
        If there is no .code  attribute, different options are there
        depending on the value of this parameter
        * None : throw an exception when no short code is found (default)
        * 'long' : keep the residue's long name, i.e. do nothing
        * 'c': any alphabetic character, as long as it is of len=1
        * 0 : the first alphabetic character in the residue's name

    keep_index : bool, default is False
        If True return "Y30" for "TRP30", instead of returning just "Y"

    Returns
    -------
    code: str
        A string representing this AA using the short code

    """

    if isinstance(AA,str):
        name = name_from_AA(AA)
        res = '%s%u' % ([name if len(name) == 1 else _AMINO_ACID_CODES.get(name)][0], int_from_AA_code(AA))
    else:
        res = '%s%u'%(AA.code,AA.resSeq)

    #print(res, AA, substitute_fail)
    if "none" in str(res).lower():
        if substitute_fail is None:
            raise KeyError("There is no short version for your input %s (%s)"%(AA, type(AA)))
        elif isinstance(substitute_fail,str):
            if substitute_fail.lower()=="long":
                res = str(AA)
            elif len(substitute_fail)==1:
                res = '%s%u'%(substitute_fail,int_from_AA_code(str(AA)))
            else:
                raise ValueError("Cannot understand your input. Please refer to the docs.")
        elif isinstance(substitute_fail,int) and substitute_fail==0:
            res = '%s%u' % (str(AA)[0], int_from_AA_code(str(AA)))
        else:
            raise ValueError("Cannot understand your input. Please refer to the docs.")

    if keep_index:
        return res
    else:
        return name_from_AA(res)

def atom_type(aa, no_BB_no_SC='X'):
    r"""
    Return a string BB or SC for backbone or sidechain atom.

    Parameters
    ----------
    aa : :obj:`mdtraj.core.topology.Atom` object
    no_BB_no_SC : str, default is X
        Return this string if :obj:`aa` isn't either BB or SC

    Returns
    -------
    aatype : str

    """
    if aa.is_backbone:
        return 'BB'
    elif aa.is_sidechain:
        return 'SC'
    else:
        return no_BB_no_SC

def parse_and_list_AAs_input(AAs, top, map_conlab=None):
    r"""Helper method to print information regarding AA descriptors

    Parameters
    ----------
    AAs : None or str
        CSVs of AA descriptors, e.g.
        'GLU30,GLU*,GDP', anything that :obj:`find_AA` can read
        How AAs are being described
    top : :obj:`~mdtraj.Topology`
        Topology where the AAs live
    map_conlab : dict, list or array, default is None
        maps residue indices to consensus labels

    Returns
    -------

    """
    if str(AAs).lower()!="none":
        AAs = [aa.strip(" ") for aa in AAs.split(",")]
        for aa in AAs:
            cands = find_AA(aa, top)
            if len(cands) == 0:
                print("No %s found in the input topology" % aa)
            else:
                for idx in cands :
                    rr = top.residue(idx)
                    if map_conlab is not None:
                        print(idx, rr, map_conlab[idx])
                    else:
                        print(idx,rr)
        print()

def find_CA(res, CA_name="CA", CA_dict=None):
    r""" Return the CA atom (or something equivalent) for this residue


    Parameters
    ----------
    res : :obj:`mdtraj.Residue` object

    CA_name : str, default is "CA"
        The name by which you identify the CA.
        This overrules anything that's parsed
        in the :obj:`CA_dict`, i.e. if the
        residue you are passing has both
        an atom "CA" and an entry
        in the CA_dict, the "CA" atom will
        be returned.

    CA_dict : dict, default is None
        You can provide a dictionary keyed
        with residue names and valued
        with strings that identify a "CA"-equivalent
        atom (e.g. in ligands)
        If None, the default :obj:`_CA_rules` are used:
        _CA_rules = {"GDP": "C1", "P0G":"C12"}

    """

    if CA_dict is None:
        CA_dict = _CA_rules

    CA_atom = list(res.atoms_by_name(CA_name))
    if len(CA_atom) == 1:
        pass
    elif len(CA_atom) == 0:
        if res.n_atoms == 1:
            return list(res.atoms)[0]
        else:
            try:
                CA_atom = list(res.atoms_by_name(CA_dict[res.name]))
                assert len(CA_atom) == 1
            except KeyError:
                raise NotImplementedError(
                    "This method does not know what the 'CA' of a %s is. Known 'CA'-equivalents are %s" % (res, CA_dict.keys()))
    else:
        raise ValueError("More than one CA atom for %s:%s"%(res,CA_atom))
    return CA_atom[0]


_CA_rules = {"GDP": "C1", "P0G":"C12"}

_AA_types = {"positive": "ARG HIS LYS",
             "negative": "ASP GLU",
             "polar": "SER THR ASN GLN",
             "special": "CYS GLY PRO",
             "hydrophobic": "ALA ILE LEU MET PHE TRP TYR VAL"}
for _key in _AA_types.keys():
    # the str().split(.) is an ugly hack for building the docs, smh the dict gets turned into a named tuple
    _AA_types[_key] += ' '+" ".join([str(_AMINO_ACID_CODES[_AA]).split(".")[-1] for _AA in _AA_types[_key].split()])
_res2restype = {aa:key for key, val in _AA_types.items() for aa in val.split()}
_res2restype

def AAtype(res,
           return_color=False,
           typecolors={"positive": "blue",
                       "negative": "red",
                       "polar": "green",
                       "special": "gray",
                       "hydrophobic": "gray",
                       "NA": "purple"}):
    r"""
    Residue types, optionally color coded

    The types are:
    * "positive": "ARG HIS LYS",
    * "negative": "ASP GLU",
    * "polar": "SER THR ASN GLN",
    * "special": "CYS GLY PRO",
    * "hydrophobic": "ALA ILE LEU MET PHE TRP TYR VAL"

    Parameters
    ----------
    res : str or :obj:`~mdtraj.core.topology.Residue`
    return_color : bool, default is False
        Return the color associated
        with the type (positive:blue, negative:red, etc)
        rather than type itself
    typecolors : dict
        The map of types to colors

    Returns
    -------

    rtype : str
        Either the type or the color
    """
    if isinstance(res,str):
        key = res
    else:
        key = res.name

    rtype = _res2restype.get(key,"NA")
    if return_color:
        return typecolors[rtype]
    else:
        return rtype

def residue_line(item_desc, residue, frag_idx,
                 consensus_maps=None,
                 fragment_names=None,
                 table=False):
    r"""Return a string that describes the residue

    Can be used justo to inform or to help dis-ambiguating:
    0.0)        GLU10 in fragment 0 with residue index  6 (CGN: G.HN.27)
    ...
    1.0)        GLU10 in fragment 1 with residue index 363


    Parameters
    ----------
    item_desc : str
        Description for the item of the list,
        "1.0" or "3.2"
    residue : :obj:`~mdtraj.core.Residue`
    frag_idx : int
        Fragment index
    fragment_names : list, default is None
        Fragment names
    consensus_maps : dict of indexables, default is None
        Dictionary of dictionaries. Lower-level dicts are keyed
        with residue indices and valued with additional residue names.
        Higher-level keys can be whatever. Use case is e.g. if "R131"
        needs to be disambiguated bc. it pops up in many fragments.
        You can pass {"GPCR":{895:"3.50", ...} here and that label
        will be displayed next to the residue.
    table : bool, default is False
        Assume a header has been aready printed
        out and print the line with the
        inline tags

    Returns
    -------
    istr : str
        An informative string about this residue, that
        can be used to dis-ambiguate via the unique
        item descriptor, e.g:
        3.1)       GLU122 in fragment 3 with residue index 852 (: 3.41)

    """
    res_idx = residue.index

    fragname = " "
    if fragment_names is not None:
        fragname = ' (%s) ' % fragment_names[frag_idx]

    if not table:
        istr = '%-6s %10s in fragment %u%swith residue index %2u' % (item_desc + ')', residue, frag_idx, fragname, res_idx)
        if consensus_maps is not None:
            extra = ''
            for key1, val1 in consensus_maps.items():
                try:
                    jstr = val1[res_idx]
                    if jstr is not None:
                        extra += '%s: %s ' % (key1, val1[res_idx])
                except (KeyError,IndexError):
                    pass
            if len(extra) > 0:
                istr = istr + ' (%s)' % extra.rstrip(" ")
    else:
        add_dicts = []
        if consensus_maps is not None:
            for key in ["GPCR","CGN"]:
                add_dicts.append(_try_double_indexing(consensus_maps, key, res_idx))

        istr = '%10s  %10u  %10u %10u %10s %10s' % (residue, res_idx,
                                                    frag_idx,
                                                    residue.resSeq,
                                                    add_dicts[0], add_dicts[1])
    return istr

def _try_double_indexing(indexable, idx1, idx2):
    try:
        return indexable[idx1][idx2]
    except (KeyError, IndexError,TypeError):
        return None


def top2lsd(top, substitute_fail="X",
            extra_columns=None):
    r"""
    Return a list of per-residue attributes as dictionaries

    Use :obj:`~pandas.DataFrame` on the return value for a nice table

    Parameters
    ----------
    top : :obj:`~mdtraj.Topology`
    substitute_fail : str, None, int, default is "X"
        If there is no .code  attribute, different options are there
        depending on the value of this parameter
         * None : throw an exception when no short code is found (default)
         * 'long' : keep the residue's long name, i.e. do nothing
         * 'c': any alphabetic character, as long as it is of len=1
         * 0 : the first alphabetic character in the residue's name
    extra_columns : dictionary of indexables
        Any other column you want to
        include in the :obj:`~pandas.DataFrame`

    Returns
    -------
    df : :obj:`~pandas.DataFrame`
    """
    list_of_dicts = []
    for rr in top.residues:
        rdict = {"residue": str(rr),
                 "index": rr.index,
                 "name": rr.name,
                 "resSeq": rr.resSeq,
                 "code": shorten_AA(rr, substitute_fail=substitute_fail, keep_index=False),
                 "short": shorten_AA(rr, substitute_fail=substitute_fail, keep_index=True)}
        if extra_columns is not None:
            for key, val in extra_columns.items():
                try:
                    rdict[key] = val[rr.index]
                except (KeyError, IndexError):
                    rdict[key] = None
        list_of_dicts.append(rdict)

    return list_of_dicts

def find_AA(AA_pattern, top,
            extra_columns=None,
            return_df=False):
    r"""

    Residue matching with UNIX-shell patterns

    Similar to the shell command "ls",
    using posix-style wildcards like
    shown in the examples or here:
    https://docs.python.org/3/library/fnmatch.html

    Any other attribute that's passed
    as :obj:`extra_columns` will be
    matched as explained below, e.g.
    "3.50" to get one residue in
    the GPCR-nomenclature or "3.*"
    to get the whole TM-helix 3

    The examples use '*' as wildcard,
    but '?' (as in 'ls') also works

    Examples
    --------
        * 'PRO' : returns all PROs, matching
          via the attribute "name"
        * 'P'   : returns all PROs, matching
          via the attribute "code"
        * 'P*'  : returns all PROs,PHEs and
          any other residue that starts with "P",
          either in "name" or in "code"
        * 'PRO39' : returns PRO39, matching
          via full residue name (long)
        * 'P39'  : returns PRO39, matching
          via full residue name (short)
        * 'PRO3*' : returns all PROs
          with sequence indices that start
          with 3, e.g. 'PRO39, PRO323, PRO330' etc
        * '3' : returns all residues with
          sequence indices 3
        * '3*' : returns all residues with
          sequence indices that start with 3


    Parameters
    ----------
    AA_patt : str or int
    top : :obj:`~mdtraj.Topology`
    return_df : bool, default is False
        Return the full :obj:`~pandas.DataFrame`
        of the matching residues

    Returns
    -------
    AAs : list or :obj:`~pandas.DataFrame`
        List of serial residue indices, s.t.
        top.residue(idx) would return the wanted residue.
        With :obj:`return_df`, you can get the
        full :obj:`~pandas.DataFrame` of the
        matching residues.
    """

    lsd = top2lsd(top, substitute_fail="X", extra_columns=extra_columns)

    idxs = [ii for ii, idict in enumerate(lsd) if _fn_filter([str(val) for key, val in idict.items() if key!="index"], str(AA_pattern))]

    if return_df:
        return _DF([lsd[ii] for ii in idxs])
    else:
        return idxs


def _ls_AA_in_df(AA_patt, df):
    r""" Same as find_AA but using dataframe syntax...between 10 and 100 times slower (200mus to 20ms)"""
    from fnmatch import fnmatch as _fnmatch
    _AA = str(AA_patt)
    match = lambda val : _fnmatch(val,_AA)
    idxs = _np.flatnonzero(df.applymap(lambda val: str(val)).applymap(match).values.any(1)).tolist()
    return idxs

def get_SS(SS,top=None):
    r"""
    Try to guess what type of input for secondary-structure computation the user wants, and compute it

    Parameters
    ----------
    SS : secondary structure information
        Can be many things:
        * triple of ints (CP_idx, traj_idx, frame_idx)
          Nothing happens, the tuple is returned as
          is and handled externally by the :obj:`ContactGroup`
          that called this method.
          Tuple representing a ContactPair, trajectory
          See the docs there for more info
        * True
          same as [0,0,0]
        * None or False
          Do nothing
        * :obj:`mdtraj.Trajectory`
          Use this geometry to compute the SS
        * string
          Path to a filename, of which only
          the first frame will be read. The
          SS will be computed from there.
          The file will be tried to read
          first without topology information
          (e.g. .pdb, .gro, .h5 will work),
          and when this fails, the :obj:`top`
          will be passed (e.g. .xtc, .dcd)
        * array_like
          Use the SS from here, s.t.ss_inf[idx]
          gives the SS-info for the residue
          with that idx
    top : :obj:`~mdtraj.Topology`, default is None

    Returns
    -------
    from_tuple : bool
        Whether the infor should be gotten from
        a tuple or not
    ss_array : np.ndarray or None
    """
    from_tuple = False
    ss_array = None
    if SS is None or (isinstance(SS, bool) and not SS):
        pass
    elif isinstance(SS, _md.Trajectory):
        ss_array = _md.compute_dssp(SS[0], simplified=True)[0]
    elif isinstance(SS, str):
        try:
            ss_array = _md.compute_dssp(_md.load(SS, frame=0), simplified=True)[0]
        except (OSError, ValueError) as e:
            ss_array = _md.compute_dssp(_md.load(SS, top=top, frame=0), simplified=True)[0]
    elif SS is True:
        from_tuple = (0, 0, 0)
    elif len(SS) == 3:
        from_tuple = SS
    else:
        ss_array = SS

    return from_tuple, ss_array


r"""
Get and manipulate consensus nomenclature for GPCRs, G-proteins, and Kinases.

Uses local files and/or accesses the following databases and their public APIs

 * `GPCRdb <https://gpcrdb.org/>`_
 * `Common Gα Numbering (CGN) scheme <https://www.mrc-lmb.cam.ac.uk/CGN/>`_
 * `KLIFS - The structural kinase database <https://klifs.net/>`_
 * `RCSB PDB rcsb.org <https://rcsb.org/>`_
 * `UniProt Knowledgebase <https://www.uniprot.org/>`_

Please see the individual documentation of the `Labeler Classes <Classes>`_ for further references and cite them
whenever you use these nomenclature schemes in your final publication.

Additionally, use :obj:`mdciao.nomenclature.references` anytime to get more info.

.. currentmodule:: mdciao.nomenclature

Classes
=======

.. autosummary::
    :toctree: generated

    LabelerGPCR
    LabelerCGN
    LabelerKLIFS
    AlignerConsensus
    Literature

Functions
=========

.. autosummary::
    :nosignatures:
    :toctree: generated/

    guess_by_nomenclature
    guess_nomenclature_fragments
    references

References
==========
These are some of the references most relevant to this module:

* GPCRdb and naming schemes therein

 * Kooistra, A. J., Mordalski, S., Pándy-Szekeres, G., Esguerra, M., Mamyrbekov, A., Munk, C., … Gloriam, D. E. (2021). GPCRdb in 2021: Integrating GPCR sequence, structure and function. Nucleic Acids Research, 49(D1), D335–D343. https://doi.org/10.1093/nar/gkaa1080

 * Isberg, V., De Graaf, C., Bortolato, A., Cherezov, V., Katritch, V., Marshall, F. H., … Gloriam, D. E. (2015). Generic GPCR residue numbers - Aligning topology maps while minding the gaps. Trends in Pharmacological Sciences, 36(1), 22–31. https://doi.org/10.1016/j.tips.2014.11.001

 * Isberg, V., Mordalski, S., Munk, C., Rataj, K., Harpsøe, K., Hauser, A. S., … Gloriam, D. E. (2016). GPCRdb: An information system for G protein-coupled receptors. Nucleic Acids Research, 44(D1), D356–D364. https://doi.org/10.1093/nar/gkv1178

* Further GPCR naming schemes

 * Ballesteros, J. A., & Weinstein, H. (1995). Integrated methods for the construction of three-dimensional models and computational probing of structure-function relations in G protein-coupled receptors. Methods in Neurosciences, 25(C), 366–428. https://doi.org/10.1016/S1043-9471(05)80049-7

 * Wu, H., Wang, C., Gregory, K. J., Han, G. W., Cho, H. P., Xia, Y., … Stevens, R. C. (2014). Structure of a class C GPCR metabotropic glutamate receptor 1 bound to an allosteric modulator. Science, 344(6179), 58–64. https://doi.org/10.1126/science.1249489

 * Pin, J. P., Galvez, T., & Prézeau, L. (2003). Evolution, structure, and activation mechanism of family 3/C G-protein-coupled receptors. Pharmacology and Therapeutics, 98(3), 325–354. https://doi.org/10.1016/S0163-7258(03)00038-X

 * Wootten, D., Simms, J., Miller, L. J., Christopoulos, A., & Sexton, P. M. (2013). Polar transmembrane interactions drive formation of ligand-specific and signal pathway-biased family B G protein-coupled receptor conformations. Proceedings of the National Academy of Sciences of the United States of America, 110(13), 5211–5216. https://doi.org/10.1073/pnas.1221585110

 * Oliveira, L., Paiva, A. C. M., & Vriend, G. (1993). A common motif in G-protein-coupled seven transmembrane helix receptors. Journal of Computer-Aided Molecular Design, 7(6), 649–658. https://doi.org/10.1007/BF00125323

 * Schwartz, T. W., Gether, U., Schambye, H. T., & Hjorth, S. A. (1995). Molecular mechanism of action of non-peptide ligands for peptide receptors. Current Pharmaceutical Design, 1, 325–342.

 * Schwartz, T. W. (1994). Locating ligand-binding sites in 7tm receptors by protein engineering. Current Opinion in Biotechnology, 5(4), 434–444. https://doi.org/10.1016/0958-1669(94)90054-X

 * Baldwin, J. M. (1993). The probable arrangement of the helices in G protein-coupled receptors. The EMBO Journal, 12(4), 1693–1703. https://doi.org/10.1002/J.1460-2075.1993.TB05814.X

 * Baldwin, J. M., Schertler, G. F. X., & Unger, V. M. (1997). An alpha-carbon template for the transmembrane helices in the rhodopsin family of G-protein-coupled receptors. Journal of Molecular Biology, 272(1), 144–164. https://doi.org/10.1006/jmbi.1997.1240

* CGN naming scheme

 * Flock, T., Ravarani, C. N. J., Sun, D., Venkatakrishnan, A. J., Kayikci, M., Tate, C. G., … Babu, M. M. (2015). Universal allosteric mechanism for Gα activation by GPCRs. Nature 2015 524:7564, 524(7564), 173–179. https://doi.org/10.1038/nature14663

* KLIFS 85 ligand binding site residues of kinases

 * Van Linden, O. P. J., Kooistra, A. J., Leurs, R., De Esch, I. J. P., & De Graaf, C. (2014). KLIFS: A knowledge-based structural database to navigate kinase-ligand interaction space. Journal of Medicinal Chemistry, 57(2), 249–277. https://doi.org/10.1021/JM400378W
 * Kooistra, A. J., Kanev, G. K., Van Linden, O. P. J., Leurs, R., De Esch, I. J. P., & De Graaf, C. (2016). KLIFS: a structural kinase-ligand interaction database. Nucleic Acids Research, 44(D1), D365–D371. https://doi.org/10.1093/NAR/GKV1082
 * Kanev, G. K., de Graaf, C., Westerman, B. A., de Esch, I. J. P., & Kooistra, A. J. (2021). KLIFS: an overhaul after the first 5 years of supporting kinase research. Nucleic Acids Research, 49(D1), D562–D569. https://doi.org/10.1093/NAR/GKAA895


* PDB

 * Berman, H. M., Westbrook, J., Feng, Z., Gilliland, G., Bhat, T. N., Weissig, H., … Bourne, P. E. (2000, January 1). The Protein Data Bank. Nucleic Acids Research. Oxford Academic. https://doi.org/10.1093/nar/28.1.235

* UniProt

 * Bateman, A., Martin, M. J., Orchard, S., Magrane, M., Agivetova, R., Ahmad, S., … Zhang, J. (2021). UniProt: the universal protein knowledgebase in 2021. Nucleic Acids Research, 49(D1), D480–D489. https://doi.org/10.1093/NAR/GKAA1100




"""
from .nomenclature import *


#!/usr/bin/env python3
##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################
from mdciao.parsers import parser_for_compare_neighborhoods
from mdciao.cli import compare
# This was originally to be able to produce plots in headless mode, but it's failing in the tests
# relevant:
# * http://omz-software.com/pythonista/matplotlib/users/shell.html
#   "With the TkAgg backend, which uses the Tkinter user interface toolkit, you can use matplotlib from an arbitrary non-gui python shell"
# * https://github.com/gph82/mdciao/runs/6352808503?check_suite_focus=true
#   raise ImportError(', "ImportError: Cannot load backend 'TkAgg' which requires the 'tk' interactive framework, as 'headless' is currently running"
#import matplotlib
#matplotlib.use('TkAgg')
# Get and instantiate parser
parser = parser_for_compare_neighborhoods()
a  = parser.parse_args()
nf = len(a.files)
if a.keys is not None:
    assert len(a.keys.split(","))==nf, "Mismatch number of files vs number of keys %u vs %u"%(nf,len(a.keys.split()))
    keys = a.keys.split(",")
    file_dict = {key:val for key, val in zip(keys, a.files)}
else:
    file_dict = a.files
a.colors = a.colors.split(",")
b = {key:getattr(a,key) for key in dir(a) if not key.startswith("_")}
for key in ["files", "mutations", "keys","output_desc"]:
    b.pop(key)
#b["figsize"]=None
b["mutations_dict"] = {}

if a.mutations is not None:
    for pair in a.mutations.split(","):
        key, val = pair.split(":")
        b["mutations_dict"][key.replace(" ","")]=val.replace(" ","")

myfig, freqs, posret = compare(file_dict,
                           output_desc=a.output_desc,
                           **b,
                           )



r"""

Functions for and around sequence alignment

The alignment takes only in one place (:obj:`my_bioalign`),
the rest of functions either prepare the alignment or produce
other objects derived from it (DataFrames, dictionaries,
maps between topologies etc)

.. currentmodule:: mdciao.utils.sequence


Functions
=========

.. autosummary::
   :toctree: generated/

"""
import numpy as _np
from pandas import DataFrame as _DF
from Bio.pairwise2 import align as _Bioalign
from .lists import contiguous_ranges as _cranges
import pandas as _pd
from IPython.display import display as _display


# See "Define original properties" https://pandas.pydata.org/pandas-docs/stable/development/extending.html#define-original-properties
class _ADF(_DF):
    r"""
    Sub-class of an :obj:`~pandas.DataFrame` to include the alignment_score as metadata.

    It can be then accessed via self.alignment_score and is preserved downstream

    Check https://pandas.pydata.org/pandas-docs/stable/development/extending.html#define-original-properties
    for more info
    """

    # normal properties
    _metadata = ["alignment_score"]

    @property
    def _constructor(self):
        return _ADF


class AlignmentDataFrame(_ADF):
    r"""
    Sub-class of an :obj:`~pandas.DataFrame` to include the alignment_score as metadata.

    Simply pass it as argument ' alignment_score=1' and it:
     * can be then accessed via self.alignment_score and
     * it is preserved downstream after operating on the df

    Check https://pandas.pydata.org/pandas-docs/stable/development/extending.html#define-original-properties
    for more info
    """

    def __init__(self,*args,**kwargs):
        alignment_score = kwargs.get("alignment_score")
        if alignment_score is not None:
            kwargs.pop("alignment_score")
        super().__init__(*args,**kwargs)
        self.alignment_score = alignment_score

def print_verbose_dataframe(df):
    r"""
    Print the full dataframe no matter how big

    Parameters
    ----------
    df

    Returns
    -------

    """
    rows, columns = df.shape
    with _pd.option_context('display.max_rows', rows,
                            'display.max_columns', columns,
                            'display.width', 1000):
        _display(df)

def top2seq(top, replacement_letter="X"):
    r"""
    Return the AA sequence of :obj:`top` as a string

    Parameters
    ----------
    top : :obj:`mdtraj.Topology`
    replacement_letter : str, default is "X"
        If the AA has no one-letter-code,
        return this letter instead has to be a str of len(1)

    Returns
    -------
    seq : str of len top.n_residues
    """
    assert len(replacement_letter)==1

    return ''.join([str(rr.code).replace("None",replacement_letter) for rr in top.residues])

def my_bioalign(seq1, seq2,
                method="globalms",
                argstuple=(1,0,-1,-.05),
                kwargs = {"penalize_end_gaps":False}):
    r"""
    Align two sequences using a method of :obj:`Bioalign`

    Note
    ----
    This is a one-liner wrapper around whatever method
    of :obj:`Bioalign` has been chosen, typically
    pairwise2.align.globalms

    The intention is to only use *this* method throughout
    mdciao, and change *here* any alignment parameters s.t.
    alignment is done using *always* the same parameters.

    The exposed arguments `method` and `argstuple`
    are there for future development but will raise
    NotImplementedErrors if changed.

    See https://biopython.org/DIST/docs/api/Bio.pairwise2-module.html
    for more info

    Parameters
    ----------
    seq1 : str, any length
    seq2 : str, any length
    method : str, default is "globalms"
    argstuple : tuple, default is (-1,0,-1,-.05)
        The tuple controlling penalties for:
        * matches
        * mismatches
        * opening a gap
        * extending the gap

    Returns
    -------
    alignments : list
        A list of tuples, each containing seq1,seq2,score.
        See https://biopython.org/DIST/docs/api/Bio.pairwise2-module.html
        for more info


    """
    # This is to be able to raise the NotImplemented but also to hard-code the only allowEd method here
    allowed_method="globalms"
    allowed_tuple = (1, 0, -1, -.05)
    if method!=allowed_method:
        raise (NotImplementedError("At the moment only %s is "
                                   "allowed as alignment method"%method))


    if tuple(argstuple) == tuple(allowed_tuple):
        return getattr(_Bioalign, method)(seq1, seq2, *argstuple, **kwargs)
    else:
        raise NotImplementedError("At the moment only %s is "
                                   "allowed as argument tuple, got"
                                   "instead %s"%(str(allowed_tuple),
                                                    str(argstuple)))



def alignment_result_to_list_of_dicts(ialg,
                                      seq_0_res_idxs,
                                      seq_1_res_idxs,
                                      topology_0=None,
                                      topology_1=None,
                                      key_AA_code_seq_0="AA_0",
                                      key_AA_code_seq_1="AA_1",
                                      key_resSeq_seq_0="resSeq_0",
                                      key_idx_seq_0="idx_0",
                                      key_idx_seq_1="idx_1",
                                      key_full_resname_seq_0='fullname_0',
                                      key_full_resname_seq_1='fullname_1',
                                      verbose=False,
                                      ):
    r"""
    Input an alignment result (:obj:`ialg`) and return it as
    a list of per-residue dictionaries with other complementary keys.

    This list of dictionaries is very suitable for further operations
    with :obj:`pandas.DataFrame`.

    TODO
    ----
    Decide whether we need key_resSeq_1 or not

    Parameters
    ----------
    ialg: list
        list with four entries, see obj:`my_bioalign`
        and https://biopython.org/DIST/docs/api/Bio.pairwise2-module.html
        for more info
    topology_0: :obj:`mdtraj.Topology` object
    seq_0_res_idxs:
        Zero-indexed residue indices of whatever was in seq_0
    seq_1_res_idxs:
        Zero-indexed residue indices of whatever was in seq_1
    key_AA_code_seq_0 : str, default is AA_0
        The key under which the residues one-letter code will
        be shown (=the column title in a :obj:`DataFrame`
    key_AA_code_seq_1 : str, default is AA_1
    key_resSeq_seq_0 : str, default is resSeq_0
    key_idx_seq_1
    key_full_resname_seq_0
    seq_0_res_idxs
    verbose: bool, default is False

    Returns
    -------
    alignment_dict : dictionary
        A dictionary containing the aligned sequences with annotated with different information

    """
    # Unpack the alignment
    top_0_seq, top_1_seq = ialg[0], ialg[1]
    # Some sanity checks
    assert len(top_0_seq) == len(top_1_seq)

    # Do we have the right indices?
    assert len(seq_1_res_idxs)==len(''.join([ii for ii in top_1_seq if ii.isalpha()]))
    assert len(seq_0_res_idxs)==len(''.join([ii for ii in top_0_seq if ii.isalpha()]))

    # Create needed iterators
    seq_1_res_idxs_iterator = iter(seq_1_res_idxs)
    idx_seq_0_iterator = iter(seq_0_res_idxs)

    top_0_resSeq_iterator = None
    resname_top_0_iterator = None
    if topology_0 is not None:
        top_0_resSeq_iterator = iter([topology_0.residue(ii).resSeq for ii in seq_0_res_idxs])
        resname_top_0_iterator = iter([str(topology_0.residue(ii)) for ii in seq_0_res_idxs])

    resname_top_1_iterator = None
    if topology_1 is not None:
        resname_top_1_iterator = iter([str(topology_1.residue(ii)) for ii in seq_1_res_idxs])

    alignment_dict = []
    for rt, rr in zip(top_0_seq, top_1_seq):
        alignment_dict.append({key_AA_code_seq_0: rt,
                               key_AA_code_seq_1: rr,
                               key_resSeq_seq_0: '~',
                               key_full_resname_seq_0: '~',
                               key_full_resname_seq_1: '~',
                               key_idx_seq_1: '~',
                               key_idx_seq_0: '~'})

        if rt.isalpha():
            if topology_0 is not None:
                alignment_dict[-1][key_full_resname_seq_0] = next(resname_top_0_iterator)
                alignment_dict[-1][key_resSeq_seq_0] = next(top_0_resSeq_iterator)
            alignment_dict[-1][key_idx_seq_0] = next(idx_seq_0_iterator)

        if rr.isalpha():
            alignment_dict[-1][key_idx_seq_1] = next(seq_1_res_idxs_iterator)
            if topology_1 is not None:
                alignment_dict[-1][key_full_resname_seq_1] = next(resname_top_1_iterator)

    # Add a field for matching vs nonmatching AAs
    for idict in alignment_dict:
        idict["match"] = False
        if idict[key_AA_code_seq_0]==idict[key_AA_code_seq_1]:
            idict["match"]=True

    if verbose:
        print("\nAlignment:")
        order = [key_idx_seq_1, key_AA_code_seq_1, key_full_resname_seq_1,
                                key_AA_code_seq_0, key_resSeq_seq_0, key_full_resname_seq_0, key_idx_seq_0, "match"]
        print_verbose_dataframe(_DF(alignment_dict)[order])

    return alignment_dict


def align_tops_or_seqs(top0, top1, substitutions=None,
                       seq_0_res_idxs=None,
                       seq_1_res_idxs=None,
                       return_DF=True,
                       verbose=False,
                       ):
    r""" Align two sequence-containing objects, i.e. strings and/or
    :obj:`~mdtraj.Topology` objects

    Returns a list of :obj:`n_best` :obj:`AlignmentDataFrame` s,
    an mdciao sub-class of a :obj:`~pandas.DataFrame`

    A list is returned because sometimes there's more than
    one alignment with the best possible score (currently it's
    limited to 10 alignments)

    Relevant methods used under the hood are :obj:`my_bioalign` and
    :obj:`alignment_result_to_list_of_dicts`, see their docs
    for more info.

    Parameters
    ----------
    top0 : :str or obj:`~mdtraj.Topology`
    top1 : :str or obj:`~mdtraj.Topology`
    substitutions : dictionary
        dictionary of patterns and replacements,
        in case some AAs of the topologies
    seq_0_res_idxs : iterable of integers, default is None
        only use these idxs for alignment in :obj:`top0`
    seq_1_res_idxs : iterable of integers, default is None
        only use these idxs for alignment in :obj:`top1`
    return_DF : bool, default is True
        If False, a list of alignment dictionaries instead
        of :obj:`AlignmentDataFrame` s will be returned
    verbose : bool, default is False

    Returns
    -------
    alignments : list of :obj:`n_best` :obj:`AlignmentDataFrame` s
        These are just normal :obj:`~pandas.DataFrames` with an extra
        attribute .alignment_score to be used downstream.
        If :obj:`return_DF` is False, it's a list of lists of dicts,
        see :obj:`alignment_result_to_list_of_dicts` for more info


    """
    if isinstance(top0, str):
        top0_seq = top0
        n_res_0 = len(top0_seq)
        top04a = None
    else:
        top0_seq = top2seq(top0)
        n_res_0 = top0.n_residues
        top04a = top0

    if isinstance(top1, str):
        top1_seq = top1
        n_res_1 = len(top1_seq)
        top14a = None
    else:
        n_res_1 = top1.n_residues
        top1_seq = top2seq(top1)
        top14a = top1

    if substitutions is not None:
        for key, val in substitutions.items():
            top0_seq = top0_seq.replace(key,val)
            top1_seq = top1_seq.replace(key,val)
            #print(key,val)

    if seq_0_res_idxs is None:
        seq_0_res_idxs=_np.arange(n_res_0, dtype=int)
    if seq_1_res_idxs is None:
        seq_1_res_idxs=_np.arange(n_res_1, dtype=int)

    top0_seq = "".join([top0_seq[ii] for ii in seq_0_res_idxs])
    top1_seq = "".join([top1_seq[ii] for ii in seq_1_res_idxs])

    alignments = my_bioalign(top0_seq, top1_seq)[:10]
    alignments = [aa for aa in alignments if aa.score == alignments[0].score]
    scores = [aa.score for aa in alignments]
    lists_of_lists_of_align_dicts = [alignment_result_to_list_of_dicts(aa,
                                                   topology_0=top04a,
                                                   seq_0_res_idxs=seq_0_res_idxs,
                                                   seq_1_res_idxs=seq_1_res_idxs,
                                                   topology_1=top14a,
                                                   verbose=verbose,
                                                   ) for aa in alignments]

    if return_DF:
        return [AlignmentDataFrame(aa, alignment_score=score) for aa, score in zip(lists_of_lists_of_align_dicts,
                                                                                  scores)]
    else:
        return lists_of_lists_of_align_dicts



def maptops(top0,
            top1,
            allow_nonmatch=False,
            ):
    r""" map residues between topologies or sequences
    via their serial indices a sequence alignment

    Parameters
    ----------
    top0 : :obj:`~mdtraj.Topology` or str
    top1:  :obj:`~mdtraj.Topology` or str
    allow_nonmatch : bool, default is False
        If true, non-matches of
        equal length will be
        considered matches

    Returns
    -------
    top0_to_top1 : dict
        top0_to_top1[10] = 20
    top1_to_top0 : dict
        top1_to_top0[20] = 10

    """
    df = align_tops_or_seqs(top0, top1,
                            return_DF=True)[0]

    return df2maps(df,allow_nonmatch=allow_nonmatch)

def df2maps(df, allow_nonmatch=True):
    r"""Map the columns "idx_0" and "idx_1" of an alignment
    (a :obj:`pandas.DataFrame`)

    Parameters
    ----------
    df : :obj:`pandas.DataFrame`
        Typically comes from  :obj:`align_tops_or_seqs`
    allow_nonmatch : bool, default is True
        Allow to map between ranges of residues that
        don't match, as long as nonmatching
        ranges are equal in length, s.t.
        A A
        A A
        B D
        B D
        C C
        C C
        maps BB to DD

    Non-matching first or last ranges will never be
    mapped

    Returns
    -------
    top0_to_top1 : dict
        top0_to_top1[10] = 20
    top1_to_top0 : dict
        top1_to_top0[20] = 10

    """
    if allow_nonmatch:
        _df = re_match_df(df)
    else:
        _df = df

    top0_to_top1 = {key: val for key, val in zip(_df[_df["match"] == True]["idx_0"].to_list(),
                                                 _df[_df["match"] == True]["idx_1"].to_list())}

    top1_to_top0 = {val:key for key, val in top0_to_top1.items()}

    return top0_to_top1, top1_to_top0

def re_match_df(df):
    r"""
    Return a copy of an alignment :obj:`pandas.Dataframe` with True 'match'-values
    for non-matching blocks that have equal length.

    For instance,
        A A True
        A A True
        B D False
        B D False
        C C True
        C C True
    gets re_matched to:
        A A True
        A A True
        B D True
        B D True
        C C True
        C C True

    The input :obj:`DataFrame` is left untouched and only a copy is returned


    Parameters
    ----------
    df : :obj:`pandas.DataFrame`
        Typically comes from  :obj:`align_tops_or_seqs`

    Returns
    -------
    _df : :obj:`pandas.DataFrame`
        A re_matched copy of :obj:`df`

    """

    match_ranges = _cranges(df["match"].values)
    _df = df.copy()
    if False in match_ranges.keys():
        for rr in match_ranges[False]:
            try:
                if all(_df.loc[[rr[0] - 1, rr[-1] + 1]]["match"]) and \
                        all(["-" not in df[key].values[rr] for key in ["AA_0",
                                                                   "AA_1"]]):  # this checks for no insertions in the alignment ("=equal length ranges")
                    _df.loc[rr, "match"] = True
            except KeyError:
                continue

    return _df

def superpose_w_CA_align(geom, ref,
                         res_indices=None,
                         ref_res_indices=None,
                         verbose=False,
                         allow_nonmatch=False):
    r"""
    Pre align on CA-atoms before calling :obj:`mdtraj.Trajectory.superpose`

    Changes :obj:`geom` in place and returns it as well

    Parameters
    ----------
    geom : :obj:`~mdtraj.Trajectory`
    ref : :obj:`~mdtraj.Trajectory`
    res_indices : iterable of ints, default is None
        Use only these indices for the sequence alignment
    ref_res_indices : iterable of ints, default is None
        Use only these indices for the sequence alignment
    allow_nonmatch : bool, default is True
        Allow to map between ranges of residues that
        don't match, as long as nonmatching
        ranges are equal in length, s.t.
        A A
        A A
        B D
        B D
        C C
        C C
        maps BB to DD

    Non-matching first or last ranges will never be
    mapped

    Returns
    -------
    geom : :obj:`~mdtraj.Trajectory`

    """
    df = align_tops_or_seqs(geom.top, ref.top,
                            seq_0_res_idxs=res_indices,
                            seq_1_res_idxs=ref_res_indices)[0]

    g2rmap, _  = df2maps(df, allow_nonmatch=allow_nonmatch)
    if verbose:
        print_verbose_dataframe(df)
    g_ats, r_ats = [],[]
    for key, val in g2rmap.items():
        g, r = None, None
        try:
            g = geom.top.residue(key).atom("CA").index
            r = ref.top.residue(val).atom("CA").index
        except KeyError:
            pass
        if None not in [g,r]:
            g_ats.append(g)
            r_ats.append(r)
    geom.superpose(ref,atom_indices=g_ats, ref_atom_indices=r_ats)
    return geom